//===-- Definition of macros to be used with assert functions -------------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#ifndef __LLVM_LIBC_MACROS_ASSERT_MACROS_H
#define __LLVM_LIBC_MACROS_ASSERT_MACROS_H

#define __STDC_VERSION_ASSERT_H__ 202311L

#endif // __LLVM_LIBC_MACROS_ASSERT_MACROS_H
