//===-- Windows Time Macros Extension -------------------------------------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#ifndef LLVM_LIBC_MACROS_WINDOWS_TIME_MACROS_EXT_H
#define LLVM_LIBC_MACROS_WINDOWS_TIME_MACROS_EXT_H

#define CLOCK_MONOTONIC 0
#define CLOCK_REALTIME 1
#define CLOCK_PROCESS_CPUTIME_ID 2
#define CLOCK_THREAD_CPUTIME_ID 3

#endif // LLVM_LIBC_MACROS_WINDOWS_TIME_MACROS_EXT_H
