//===-- Definition of macros from time.h ---------------------------------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#ifndef LLVM_LIBC_MACROS_GPU_TIME_MACROS_H
#define LLVM_LIBC_MACROS_GPU_TIME_MACROS_H

#define CLOCK_REALTIME 0
#define CLOCK_MONOTONIC 1

#define CLOCKS_PER_SEC 1000000

#endif // LLVM_LIBC_MACROS_GPU_TIME_MACROS_H
