//===----------------------------------------------------------------------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//
//
// Defines cir type interfaces.
//
//===----------------------------------------------------------------------===//

#include "clang/CIR/Interfaces/CIRTypeInterfaces.h"

using namespace cir;

/// Include the generated interfaces.
#include "clang/CIR/Interfaces/CIRTypeInterfaces.cpp.inc"
