//===-- Definition of type __dl_iterate_phdr_callback_t ------------------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===---------------------------------------------------------------------===//

#ifndef LLVM_LIBC_TYPES___DL_ITERATE_PHDR_CALLBACK_T_H
#define LLVM_LIBC_TYPES___DL_ITERATE_PHDR_CALLBACK_T_H

#include "size_t.h"

struct dl_phdr_info;

typedef int (*__dl_iterate_phdr_callback_t)(struct dl_phdr_info *, size_t,
                                            void *);

#endif // LLVM_LIBC_TYPES___DL_ITERATE_PHDR_CALLBACK_T_H
