//===-- Definition of EFI_INTERFACE_TYPE type -----------------------------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#ifndef LLVM_LIBC_TYPES_EFI_INTERFACE_TYPE_H
#define LLVM_LIBC_TYPES_EFI_INTERFACE_TYPE_H

typedef enum {
  EFI_NATIVE_INTERFACE,
} EFI_INTERFACE_TYPE;

#endif // LLVM_LIBC_TYPES_EFI_INTERFACE_TYPE_H
